unit uDirSync;
{$INCLUDE 'DirSync3.inc'}

interface

implementation

uses
  Windows, SysUtils, Classes,
  {$INCLUDE 'DelphiFarUnits.inc'} ,
  {$IFDEF DEBUG_LOGS} uLog, {$ENDIF}
  uSynchronizeDirectories,
  uSystem, uDirSyncConsts, uMessages, uDiscoverDifferences,
  uOptions, uOptionsDialog, uEditorTasks;

type
  TDirSyncPlugin = class(TFarPlugin)
  private
    fPluginInfoAvailable: boolean;
    fPluginInfo: TPluginInfo;
  protected
    function PluginID: TGUID; override;
    function GetPluginTitle(out Description, Author: string): string; override;
    function GetPluginVersion: TVersionInfo; override;
    function GetPluginInfo(out Info: TPluginInfo): boolean; override;
    function Execute(OpenFrom: DWORD; GUID: PGUID; Data: TIntPtr): THandle; override;
    function GetCommonString(ID: TFarPlugin.TCommonString): string; override;
    procedure DoLog(const Msg: string); override;
    function PluginConfig(GUID: PGUID; MenuItem: TIntPtr): boolean; override;
  public
    constructor Create;
    destructor Destroy; override;
  end;

const
  PLUGIN_TITLE = 'DirSync3';
  PLUGIN_DESCRIPTION = 'Synchronize directories';
  PLUGIN_AUTHOR = 'Pepak, http://www.pepak.net';

{ TDirSyncPlugin }

constructor TDirSyncPlugin.Create;
begin
  inherited;
  fPluginInfoAvailable := False;
end;

destructor TDirSyncPlugin.Destroy;
begin
  if fPluginInfoAvailable then begin
    DeletePluginInfo(fPluginInfo);
  end;
  inherited;
end;

procedure TDirSyncPlugin.DoLog(const Msg: string);
begin
  {$IFDEF DEBUG_LOGS}
  uLog.Log(Msg);
  {$ENDIF}
end;

function TDirSyncPlugin.Execute(OpenFrom: DWORD; GUID: PGUID; Data: TIntPtr): THandle;
var
  Diff: TDiscoverDifferences;
  Edit: TEditorTasks;
begin
  if OpenFrom = OPEN_EDITOR then begin
    Edit := TEditorTasks.Create;
    try
      Edit.EditorID := -1; // active editor
      Result := Edit.Execute;
    finally
      FreeAndNil(Edit);
    end;
  end
  else begin
    Diff := TDiscoverDifferences.Create;
    try
      Result := Diff.Execute;
    finally
      FreeAndNil(Diff);
    end;
  end;
end;

function TDirSyncPlugin.GetCommonString(ID: TFarPlugin.TCommonString): string;
begin
  case ID of
    TFarPlugin.TCommonString.MTitle:
      Result := GetMsg(MTitle);
    TFarPlugin.TCommonString.MExceptionOccurredIn:
      Result := GetMsg(MExceptionOccurredIn);
    TFarPlugin.TCommonString.MOk:
      Result := GetMsg(MOk);
    TFarPlugin.TCommonString.MCancel:
      Result := GetMsg(MCancel);
    TFarPlugin.TCommonString.MYes:
      Result := GetMsg(MYes);
    TFarPlugin.TCommonString.MNo:
      Result := GetMsg(MNo);
    else
      Result := inherited GetCommonString(ID);
  end;
end;

function TDirSyncPlugin.GetPluginInfo(out Info: TPluginInfo): boolean;
var
  fMainMenu: TFarPlugin.TFarPluginMenu;
  fConfigMenu: TFarPlugin.TFarPluginMenu;
  fDiskMenu: TFarPlugin.TFarPluginMenu;
begin
  if not fPluginInfoAvailable then begin
    SetLength(fMainMenu, 1);
    fMainMenu[0].Title := GetMsg(MTitle);
    fMainMenu[0].ID := MENU_PANEL_GUID;
    SetLength(fConfigMenu, 1);
    fConfigMenu[0].Title := GetMsg(MTitle);
    fConfigMenu[0].ID := MENU_CONFIG_GUID;
    SetLength(fDiskMenu, 0);
    Self.fPluginInfo := CreatePluginInfo(
      PF_EDITOR,
      fMainMenu, fConfigMenu, fDiskMenu,
      ''
    );
    fPluginInfoAvailable := True;
  end;
  Info := Self.fPluginInfo;
  Result := True;
end;

function TDirSyncPlugin.GetPluginTitle(out Description, Author: string): string;
begin
  Result := PLUGIN_TITLE;
  Description := PLUGIN_DESCRIPTION;
  Author := PLUGIN_AUTHOR;
end;

function TDirSyncPlugin.GetPluginVersion: TVersionInfo;
var
  FN: string;
  VerHigh, VerLow, VerBuild: LongWord;
begin
  FN := GetModuleName(HInstance);
  if GetFileVersion(FN, VerHigh, VerLow, VerBuild) then
    Result := MakeFarVersion( {$IFDEF FAR3_UP} VerHigh, VerLow, 0, VerBuild, 0 {$ELSE} VerHigh, VerLow, VerBuild {$ENDIF} )
  else
    Result := MakeFarVersion( {$IFDEF FAR3_UP} 0, 0, 0, 0, 0 {$ELSE} 0, 0 ,0 {$ENDIF} );
end;

function TDirSyncPlugin.PluginConfig(GUID: PGUID; MenuItem: TIntPtr): boolean;
var
  Dlg: TOptionsDialog;
begin
  Result := False;
  Dlg := TOptionsDialog.Create;
  try
    Dlg.Options := TDirSyncOptions.Create;
    try
      Dlg.Options.Load;
      if Dlg.Execute then
        Result := True;
    finally
      Dlg.Options.Free;
      Dlg.Options := nil;
    end;
  finally
    FreeAndNil(Dlg);
  end;
end;

function TDirSyncPlugin.PluginID: TGUID;
begin
  Result := PLUGIN_GUID;
end;

function ProcessEditorEventW( {$IFDEF FAR3_UP} const Info: TProcessEditorEventInfo {$ELSE} Event: integer; Param: Pointer {$ENDIF} ): TIntPtr; stdcall;
begin
  Result := 0;
  {$IFDEF FAR3_UP}
  if Info.StructSize >= Sizeof(Info) then
  {$ENDIF}
    TFarEditorManaged.ProcessEditorEvent( {$IFDEF FAR3_UP} Info.Event, Info.EditorID {$ELSE} Event, -1 {$ENDIF} , Result);
end;

function ProcessEditorInputW( {$IFDEF FAR3_UP} const Info: TProcessEditorInputInfo {$ELSE} Info: PInputRecord {$ENDIF} ): TIntPtr; stdcall;

  function RunEditorTask(const EditorTask: TEditorTasks.TEditorTask): TIntPtr;
    var
      Tasks: TEditorTasks;
    begin
      //Log('Running editor task %d', [Integer(EditorTask)]);
      Tasks := TEditorTasks.Create;
      try
        Tasks.EditorID := -1;
        if Tasks.ExecuteTask(EditorTask) then begin
          Result := 1;
          TFarEditor.Refresh(-1);
          //Log('Running editor task %d -> success', [Integer(EditorTask)]);
        end
        else begin
          Result := 0;
          //Log('Running editor task %d -> failure', [Integer(EditorTask)]);
        end;
      finally
        FreeAndNil(Tasks);
      end;
    end;

var
  ShiftFlags: DWORD;
  P: PInputRecord;
begin
  Result := 0;
  {$IFDEF FAR3_UP}
  P := @Info.Rec;
  {$ELSE}
  P := Info;
  {$ENDIF}
  {$IFDEF FAR3_UP}
  if Info.StructSize >= Sizeof(Info) then
  {$ENDIF}
    if P^.EventType = KEY_EVENT then
      if P^.Event.KeyEvent.bKeyDown then begin
        ShiftFlags := P^.Event.KeyEvent.dwControlKeyState and (LEFT_ALT_PRESSED or LEFT_CTRL_PRESSED or RIGHT_ALT_PRESSED or RIGHT_CTRL_PRESSED or SHIFT_PRESSED);
        //Log('Editor input. Char=%d, VirtualKeyCode=%d, VirtualScanCode=%d, ShiftFlags = %x', [Ord(P^.Event.KeyEvent.UnicodeChar), P^.Event.KeyEvent.wVirtualKeyCode, P^.Event.KeyEvent.wVirtualScanCode, ShiftFlags]);
        if (ShiftFlags = LEFT_ALT_PRESSED) or (ShiftFlags = RIGHT_ALT_PRESSED) then begin
          if P^.Event.KeyEvent.wVirtualKeyCode = VK_F1 then
            Result := RunEditorTask(etCopyToLeft)
          else if P^.Event.KeyEvent.wVirtualKeyCode = VK_F2 then
            Result := RunEditorTask(etCopyToRight)
          else if P^.Event.KeyEvent.wVirtualKeyCode = VK_F3 then
            Result := RunEditorTask(etCompare)
          else if P^.Event.KeyEvent.wVirtualKeyCode = VK_F5 then
            Result := RunEditorTask(etFindNextToLeft)
          else if P^.Event.KeyEvent.wVirtualKeyCode = VK_F6 then
            Result := RunEditorTask(etFindNextToRight)
        end
        else if (ShiftFlags = LEFT_CTRL_PRESSED) or (ShiftFlags = RIGHT_CTRL_PRESSED) then begin
          if P^.Event.KeyEvent.wVirtualKeyCode = VK_F1 then
            Result := RunEditorTask(etEditLeft)
          else if P^.Event.KeyEvent.wVirtualKeyCode = VK_F2 then
            Result := RunEditorTask(etEditRight)
          else if P^.Event.KeyEvent.wVirtualKeyCode = VK_F3 then
            Result := RunEditorTask(etFileInfo)
        end
      end;
end;

exports
  ProcessEditorEventW,
  ProcessEditorInputW;

initialization
  TFarPlugin.InitInstance(TDirSyncPlugin.Create);

end.
